/* Copyright (C) 2002-2019 Free Software Foundation, Inc.
   Contributed by Zack Weinberg <zack@codesourcery.com>

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 3, or (at your option) any later
version.

GCC is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

Under Section 7 of GPL version 3, you are granted additional
permissions described in the GCC Runtime Library Exception, version
3.1, as published by the Free Software Foundation.

You should have received a copy of the GNU General Public License and
a copy of the GCC Runtime Library Exception along with this program;
see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
<http://www.gnu.org/licenses/>.  */

/* Threads compatibility routines for libgcc2 for VxWorks.
   These are out-of-line routines called from gthr-vxworks.h.  */

#include "tconfig.h"
#include "tsystem.h"
#include "gthr.h"

#if defined(__GTHREADS)
#include <vxWorks.h>
#ifndef __RTP__
#include <vxLib.h>
#endif
#ifndef __RTP__
#include <taskHookLib.h>
#else
# include <errno.h>
#endif

/* Init-once operation.

   This would be a clone of the implementation from gthr-solaris.h,
   except that we have a bootstrap problem - the whole point of this
   exercise is to prevent double initialization, but if two threads
   are racing with each other, once->mutex is liable to be initialized
   by both.  Then each thread will lock its own mutex, and proceed to
   call the initialization routine.

   So instead we use a bare atomic primitive (vxTas()) to handle
   mutual exclusion.  Threads losing the race then busy-wait, calling
   taskDelay() to yield the processor, until the initialization is
   completed.  Inefficient, but reliable. In RTP mode, we do not have
   have access to the vxTas primitive, instead, we use vxCas(). */

void __maybe_unlock_gthread_once_t (__gthread_once_t ** __guard)
{
  /* If done is not set, the underlying function raised an exception.  */
  if (!(*__guard)->done)
    (*__guard)->busy = 0;
}

/* Test And Set API.  For kernel mode, we have vxTas at hand.  For RTP, either
   vxCas or vxAtomicCas, depending on the VxWorks version.  */

#ifdef __RTP__

#if _VXW_MAJOR < 6 || (_VXW_MAJOR == 6 && _VXW_MINOR < 9)
#define __TAS(x) vxCas ((x), 0, 1)
#else
#define __TAS(x) vxAtomicCas ((x), 0, 1)
#endif

#else /* ! __RTP__ */

#define __TAS(x) vxTas (x)

#endif


int __gthread_once (__gthread_once_t * __guard, void (*__func) (void))
{
  if (__guard->done)
    return 0;

  while (!__TAS(&__guard->busy))
    {
      /* We wait for the task that is currently running to be over.  */
      if (__guard->done)
	return 0;
      taskDelay (1);
    }

  if (!__guard->done)
    {
#ifndef __USING_SJLJ_EXCEPTIONS__
      /* This cleanup function will be called when the variable goes out of
         scope. It is here to handle cases when __func() throws an exception.

	 We cannot use this attribute with SJLJ exceptions as the generated
	 call to Unwind_Register calls itself __gthread_once, leading to
	 an infinite recursion.  */
      __attribute__ ((cleanup (__maybe_unlock_gthread_once_t)))
	__gthread_once_t *__temp = __guard;
#endif

      __func ();
      __guard->done = 1;
    }

  /* We are no longer busy, we can process the next thread.  */
  __guard->busy = 0;
  return 0;
}

#endif /* __GTHREADS */
